(* $Id: godi_profile.ml 1025 2012-06-17 19:02:34Z gerd $ *)

open Printf
open Godi_types

type profile =
    { prof_name : string;  
      prof_comment : string;
      prof_section : string;
      prof_sources : prof_source list;
      prof_origin : string option;
      prof_options : (string * string) list;
      prof_patches : (string * string) list;
      prof_archive : bool;
    }

and prof_source =
    { prof_source : string;   (* Symbol of the source *)
      prof_enable_condition : [ `Always | `Tools ];
      prof_include : Godi_versions.dependency list;
      prof_build_sites : string list;
      prof_master_sites : string list;
    }


let ws_re = Str.regexp "[ \t\r\n]+"


let name_encode s =
  (* Encode unsafe chars as "_<hex>" *)
  let b = Buffer.create 80 in
  String.iter
    (fun c ->
       match c with
	 | '0'..'9'
	 | 'A'..'Z'
	 | 'a'..'z'
	 | '-' -> Buffer.add_char b c
	 | _ ->
	     bprintf b "_%02X" (Char.code c)
    )
    s;
  Buffer.contents b


let ndec_re = Str.regexp "_\\([0-9A-Fa-f][0-9A-Fa-f]\\)"

let name_decode s =
  Str.global_substitute
    ndec_re
    (fun s1 ->
       let u = Str.matched_group 1 s1 in
       String.make 1 (Char.chr (int_of_string ("0x" ^ String.sub u 0 2)))
    )
    s


let value_encode s =
  Netencoding.Base64.encode s


let value_decode s =
  Netencoding.Base64.decode s



let read_profile_ini filename =
  try
    let ini =
      new Godi_inifiles.inifile (Godi_file.nativename filename) in
    let name =
      ini#getval "profile" "name" in
    let comment =
      ini#getval "profile" "comment" in
    let section =
      ini#getval "profile" "section" in
    let source_ids =
      Str.split ws_re (ini#getval "profile" "sources") in
    let sources =
      List.map
	(fun sid ->
	   let sname =
	     ini#getval sid "name" in
           let senable =
             try
               match ini#getval sid "enable_condition" with
                 | "always" -> `Always
                 | "tools" -> `Tools
                 | _ -> failwith "bad param enable_condition"
             with _ -> `Always in
	   let include_l =
	     Str.split ws_re (ini#getval sid "include") in
	   { prof_source = sname;
             prof_enable_condition = senable;
	     prof_include = List.map Godi_versions.parse_dependency include_l;
	     prof_build_sites =
	       Str.split ws_re (ini#getval sid "build_sites");
	     prof_master_sites =
	       Str.split ws_re (ini#getval sid "master_sites");
	   }
	)
	source_ids in
    let origin =
      try Some(ini#getval "profile" "origin") with _ -> None in
    let option_names =
      try ini#attrs "profile_options" with _ -> [] in
    let options =
      List.map
	(fun n -> (n, ini#getval "profile_options" n))
	option_names in
    let patch_enc_names =
      try ini#attrs "profile_patches" with _ -> [] in
    let patches =
      List.map
	(fun enc_n ->
	   let n = name_decode enc_n in
	   let s = ini#getval "profile_patches" enc_n in
	   let v = value_decode s in
	   (n,v)
	)
	patch_enc_names in
    let archive =
      try ini#getval "profile" "archive" = "true" with _ -> false in
    { prof_name = name;
      prof_comment = comment;
      prof_section = section;
      prof_sources = sources;
      prof_origin = origin;
      prof_options = options;
      prof_patches = patches;
      prof_archive = archive;
    }
  with
    | err ->
	failwith ("Invalid profile: " ^ Godi_file.slashname filename ^ 
		    " - error: " ^ Netexn.to_string err)


let write_profile_ini filename p =
  Godi_file.write_file_lines  ~mode:`Text filename ["[profile]"; "name=foo" ];
  try
    let ini =
      new Godi_inifiles.inifile (Godi_file.nativename filename) in
    ini#delval "profile" "name";
    let next_id = ref 0 in
    let sources =
      List.map 
	(fun s -> 
	   let id = !next_id in
	   incr next_id;
	   ("source_" ^ string_of_int id), s
	) 
	p.prof_sources in
    ini # setval "profile" "name" p.prof_name;
    ini # setval "profile" "comment" p.prof_comment;
    ini # setval "profile" "section" p.prof_section;
    ini # setval "profile" "sources" 
      (String.concat " " (List.map fst sources));
    List.iter
      (fun (sid,s) ->
	 ini # setval sid "name" s.prof_source;
         ini # setval sid "enable_condition"
           ( match s.prof_enable_condition with
               | `Always -> "always"
               | `Tools -> "tools"
           );
	 ini # setval 
	   sid
	   "include"
	   (String.concat " "
	      (List.map Godi_versions.print_dependency s.prof_include));
	 ini # setval
	   sid "build_sites" (String.concat " " s.prof_build_sites);
	 ini # setval
	   sid "master_sites" (String.concat " " s.prof_master_sites);
      )
      sources;
    ( match p.prof_origin with
	| None -> ()
	| Some url ->
	    ini # setval "profile" "origin" url
    );
    List.iter
      (fun (n,v) ->
	 ini # setval "profile_options" n v
      )
      p.prof_options;
    List.iter
      (fun (n,v) ->
	 ini # setval "profile_patches" (name_encode n) (value_encode v)
      )
      p.prof_patches;
    ini # setval "profile" "archive" (string_of_bool p.prof_archive);
    ini # save()
  with
    | err ->
	failwith ("Invalid profile: " ^ Godi_file.slashname filename ^ 
		    " - error: " ^ Netexn.to_string err)


let write_profile_conf filename p =
  (* there is intentionally no quoting for site URLs, so that the profile
     can point to variables, e.g. build_sites = "$(GODI_STANDARD_BUILD_SITES)"
   *)
  let is_enabled src =
    match src.prof_enable_condition with
      | `Always -> true
      | `Tools -> src.prof_source = Godi_conf.tools_source in
  let lines =
    [ sprintf "# profile %s" p.prof_name;
      "# Warning: This file is generated. Do not edit here, but in godi.conf!";
      sprintf "GODI_SECTION = %s" p.prof_section;
    ]
    @
      (List.flatten
	 (List.map
	    (fun src ->
               if is_enabled src then
	         List.map
		   (fun site ->
		      sprintf "GODI_BUILD_SITES += %s" site
		   )
		   src.prof_build_sites
               else
                 []
	    )
	    p.prof_sources
	 )
      ) 
    @
      (List.flatten
	 (List.map
	    (fun src ->
               if is_enabled src then
	         List.map
		   (fun site ->
		      if site = "" || site.[0] = '$' then
		        sprintf  (* support variables *)
			  "MASTER_SITES_PROFILE += %s" site
		      else
		        sprintf
			  "MASTER_SITES_PROFILE += {%s}%s" src.prof_source site
		   )
		   src.prof_master_sites
               else
                 []
	    )
	    p.prof_sources
	 )
      ) 
    @ 
      (List.map
	 (fun (n,v) ->
	    sprintf "%s = %s" n v
	 )
	 p.prof_options
      )
  in
  Godi_file.write_file_lines ~mode:`Text filename lines


let patch_name_re =
  Str.regexp "^\\([a-zA-Z0-9_-][^/]*\\)/\
               \\([a-zA-Z0-9_-][^/]*\\)/\
               \\([a-zA-Z0-9_-][^/]*\\)$"


let write_profile_patches tree p =
  Godi_file.remove ~recurse:true tree;
  Godi_file.create_dir tree;
  List.iter
    (fun (n,v) ->
       if not (Str.string_match patch_name_re n 0) then
	 failwith ("Godi_profile.write_profile_patches: invalid patch name: " ^
		     n);
       let cat = Str.matched_group 1 n in
       let pkg = Str.matched_group 2 n in
       let file = Str.matched_group 3 n in
       let dir = Godi_file.addname(Godi_file.addname tree cat) pkg in
       Godi_file.create_dir ~recurse:true dir;
       let path = Godi_file.addname dir file in
       Godi_file.write_file ~mode:`Binary path v
    )
    p.prof_patches


let update_profile_patches tree p =
  let tree_filename = Godi_file.nativename tree in
  let files =
    Netglob.glob ~base_dir:tree_filename (`String "*/*/{file,patch}-*") in
  let patches =
    List.map
      (fun n ->
         if not (Str.string_match patch_name_re n 0) then
	   failwith
             ("Godi_profile.update_profile_patches: invalid patch name: " ^
		n);
         let path = Godi_file.addname tree n in
         let data = Godi_file.read_file ~mode:`Binary path in
         (n,data)
      )
      files in
  { p with prof_patches = patches }


let list_profiles d =
  let names =
    Godi_file.read_dir d in
  List.filter
    (fun n ->
       Godi_file.file_exists
	 (Godi_file.addname (Godi_file.addname d n) "profile.ini")
    )
    names


let pack_profile ppath tdir =
  if not (Godi_file.file_exists (Godi_file.addname ppath "profile.ini")) then
    failwith "Godi_profile.pack_profile: bad path";
  let d, b = 
    match Godi_file.dirname ppath with
      | Some d -> (d, Godi_file.basename ppath)
      | None -> failwith "Godi_profile.pack_profile: bad path" in
  let iterator =
    Pax.file_iterator_of_list
      ~prefix:d
      [b] in
  let out =
    Godi_file.addname
      (Godi_file.tmpdir_path tdir)
      (sprintf "%s.profile.tgz" b) in
  let tgz =
    Pax.create (`File out) `Gzip `Tar in
  let file = ref (Pax.next_file iterator) in
  while !file <> None do
    let f = match !file with Some f -> f | None -> assert false in
    Pax.add_element_from_file
      ~prefix:d
      tgz
      f;
    file := Pax.next_file iterator;
  done;
  Pax.close_out tgz;
  out


let slash_re = Str.regexp "/+"

let unpack_profile intgz d =
  let b = Godi_file.basename intgz in
  if not (Filename.check_suffix b ".profile.tgz") then
    failwith ("Godi_profile.unpack_profile: bad input filename: " ^ 
		Godi_file.slashname intgz);
  let n = Filename.chop_suffix b ".profile.tgz" in
  Godi_file.remove ~recurse:true (Godi_file.addname d n);
  try
    let tgz = Pax.read (`File intgz) `Gzip `Tar in
    let elem = ref (Pax.read_element tgz) in
    while !elem <> None do
      let e = match !elem with Some e -> e | None -> assert false in
      let m = Pax.get_metadata e in
      let p = m#name in
      let pl = Str.split_delim slash_re p in
      ( match pl with
	  | p0 :: _ ->
	      if p0 <> n then
		failwith "Godi_profile.unpack_profile: format error"
	  | [] ->
	      failwith "Godi_profile.unpack_profile: format error"
      );
      Pax.extract_element_as_file
	~prefix:d
	~preserve_mtime:true
	~overwrite:`Reject
	e;
      elem := Pax.read_element tgz
    done;
    Pax.postprocess_extraction tgz;
    Pax.close_in tgz
  with
    | error ->
	Godi_file.remove ~recurse:true (Godi_file.addname d n);
	raise error


let profile_match p (wp:Godi_types.wrapped_package) =
  let pkg = wp#build_pkg in
  let p_src = pkg.Godi_pkglist.pkg_source in
  let p_name = pkg.Godi_pkglist.pkg_name in
  try
    let src =
      List.find (fun s -> s.prof_source = p_src) p.prof_sources in
    let p_deps =
      List.filter
	(fun (n,_) -> n=p_name)
	src.prof_include in
    if p_deps = [] then ( (* check for *-* *)
      List.mem (Godi_versions.any_dependency "*") src.prof_include
    )
    else
      let v = wp#version in
      List.for_all
	(fun (_,cond) ->
	   Godi_versions.version_matches_condition v cond
	)
	p_deps
  with 
    | Not_found -> false
